/*---------------------------------------------------------------------------*\

Notices:

    Copyright © 2010 United States Government as represented by the
    Administrator of the National Aeronautics and Space Administration.  All
    Rights Reserved.

Disclaimers

    No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY
    OF ANY KIND, EITHER EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT
    LIMITED TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO
    SPECIFICATIONS, ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
    PARTICULAR PURPOSE, OR FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE
    SUBJECT SOFTWARE WILL BE ERROR FREE, OR ANY WARRANTY THAT DOCUMENTATION,
    IF PROVIDED, WILL CONFORM TO THE SUBJECT SOFTWARE. THIS AGREEMENT DOES
    NOT, IN ANY MANNER, CONSTITUTE AN ENDORSEMENT BY GOVERNMENT AGENCY OR ANY
    PRIOR RECIPIENT OF ANY RESULTS, RESULTING DESIGNS, HARDWARE, SOFTWARE
    PRODUCTS OR ANY OTHER APPLICATIONS RESULTING FROM USE OF THE SUBJECT
    SOFTWARE.  FURTHER, GOVERNMENT AGENCY DISCLAIMS ALL WARRANTIES AND
    LIABILITIES REGARDING THIRD-PARTY SOFTWARE, IF PRESENT IN THE ORIGINAL
    SOFTWARE, AND DISTRIBUTES IT "AS IS."

    Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS
    AGAINST THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS,
    AS WELL AS ANY PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT
    SOFTWARE RESULTS IN ANY LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES
    ARISING FROM SUCH USE, INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR
    RESULTING FROM, RECIPIENT'S USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL
    INDEMNIFY AND HOLD HARMLESS THE UNITED STATES GOVERNMENT, ITS CONTRACTORS
    AND SUBCONTRACTORS, AS WELL AS ANY PRIOR RECIPIENT, TO THE EXTENT
    PERMITTED BY LAW.  RECIPIENT'S SOLE REMEDY FOR ANY SUCH MATTER SHALL BE
    THE IMMEDIATE, UNILATERAL TERMINATION OF THIS AGREEMENT.

  ---------------------------------------------------------------------------

Application PAM

    PAM is a multi-option sequential solver implementing a generic model for
    multiphase porous materials.

Description of PAM

    Please refer to Lachaud2017_PATO for a full description of PAM.

\*---------------------------------------------------------------------------*/

#include "fvCFD.H"
#include "dynamicFvMesh.H"
#include "dynamicMotionSolverFvMesh.H"

#include "IFstream.H"
#include "IOstream.H"
#include "ISstream.H"
#include "BprimeSI.H"
#include "BLinput.H"
#include "materialProperties.H"
#include "gasProperties.H"
#include "RectangularMatrix.H"
#include "fluxFactor.H"

#include "fixedValueFvPatchField.H"
#include "fixedGradientFvPatchField.H"
#include "fixedGradientFvPatchFields.H"

#include "psiReactionThermo.H"
#include "psiChemistryModel.H"
#include "chemistrySolver.H"
#include "OFstream.H"
#include "thermoPhysicsTypes.H"
#include "basicMultiComponentMixture.H"

#include "probes.H"
#include "IOprobes.H"

#undef Log // conflict between OpenFoam and Mutation++ on the alias "Log" -> undefined here from OpenFoam.
#include <mutation++.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "fvOptions.H"

#include "regionProperties.H"
// #include "regionCouplePolyPatch.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

int main(int argc, char *argv[])
{

    #include "setRootCase.H"
    #include "createTime.H"

    regionProperties rp(runTime);

    // set the meshes and read properties for multi-materials
    #include "createFields/createSolidMeshes.H"
    #include "createFields/createPorousFields.H"         // 1 porous material
    #include "createFields/createSubStructureFields.H"   // n substructure materials
    #include "createFvOptions.H"
    #include "input/readGasProperties.H"
    #include "input/readVirginMaterialThermoTable.H"
    #include "input/readCharredMaterialThermoTable.H"
    #include "input/readBprimeTableSI.H"
    #include "input/readBoundaryConditionProperties.H"
    #include "input/readMotionProperties.H"
    #include "createFields/createStreams.H"
    #include "input/readChemistryProperties.H"
    #include "updateDerivedFields/computeData.H"
    #include "input/setFluxFactorMap.H"
    #include "output/write.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

    while (runTime.loop())
    {

        Info << "-----------" << nl << "runTime = " << runTime.timeName()
             << " s ; Time step = " << runTime.deltaT().value() << " s" << endl;

        // set the boundary conditions on the porous material surface
        #include "boundaryConditions/setPressureBC.H" // surface pressure
        #include "meshMotion/setMotion.H"             // surface motion
        #include "boundaryConditions/setEnergyBC.H"   // surface temperature

        // compute the pyrolysis rate
        Info << "Pyrolysis" << nl;
        #include "chemistry/pyrolysis.H"

        // compute gas and solid properties
        Info << "update data" << nl;
        #include "updateDerivedFields/computeData.H"

        // solve gas mass-momentum conservation equation
        Info << "gass mass" << nl;
        #include "massConservation/gMass.H"

        // solve finite-rate chemistry and species conservation
        if (finiteRate)
        {
            Info << "chemistry rates" << nl;
            #include "chemistry/chemistry.H"

            if (volAblation)
            {
                Info << "solid mass" << nl;
                #include "massConservation/sMass.H" // solve solid-mass conservation equation (in-depth ablation)
                #include "boundaryConditions/setSpeciesBC.H"
            }

            Info << "species mass" << nl;
            #include "speciesConservation/YEqn.H" // solve gas species mass-fraction conservation equation
        }

        // solve element conservation (in equilibrium mode only)
        else
        {
            if (elementConservation)
            {
                // update wall element composition (from B') and solve element conservation equation
                #include "boundaryConditions/setElementsBCgrad.H"
                #include "speciesConservation/ZEqn.H"
            }
        }

        // solve energy conservation equations
        Info << "energy" << nl;
        #include "energyConservation/energy2T.H"           // porous material
        #include "energyConservation/energySubStructure.H" // substructure materials

        // output data files
        Info << "write" << nl;
        #include "output/write.H"

        Info << "ExecutionTime = " << runTime.elapsedCpuTime() << " s"
             << "  ClockTime = " << runTime.elapsedClockTime() << " s"
             << nl << endl;

        // adaptative time step set on the max pressure gradient.
        #include "timeStep/setTimeStep_gradp.H"
    }

    Info << "End\n" << endl;

    return (0);
}


// ************************************************************************* //
