/*---------------------------------------------------------------------------*\
setEnergyBC.H

Application
    Set the temperature boundary condition on the surface patch.

Description
    There is a switch in the file 'constant/dynamicMeshDict' called:  boundaryLayerApprox.
 	- "yes" : CMA-type boundary layer approximation. It needs some simplified
 boundary layer information ('constant/tabulatedData/boundaryLayer') and a B' table 
('constant/tabulatedData/Bprime'). Numerically, a 'temperture gradient' boundary condition
 is used. It uses the 'gradTempAccess' boundary condition (in libPATO/) that is derived from
 the standard OpenFOAM 'fixedGradient' boundary condition.
	- "no" : reads the motion and the surface temperature directly from file
 'input/motionAndTemperature'. Numerically, a 'temperature' boundary condition is used.
 It uses the standard OpenFOAM 'fixedValue' boundary condition.

\*---------------------------------------------------------------------------*/

if(boundaryLayerApprox) // if boundaryLayerApprox = "yes" (see description in the header)
{	// Grab T wall
	scalarField Tw = Ta.boundaryField()[movingPatchID];
	scalarField pw = p.boundaryField()[movingPatchID];
	// Grab surface normal vectorField to the patch
	vectorField nf = 
		- mesh.Sf().boundaryField()[movingPatchID]
		/mesh.magSf().boundaryField()[movingPatchID];
	// Grab the conductivity at the wall
 	tensorField kw = k.boundaryField()[movingPatchID];
	// Projection of k on the surface normal
	scalarField kProj = nf & kw & nf; 
// Info << "kProj = " << kProj <<nl; 
	// Update the temperature gradient at the wall (explicit)
	scalarField TbcPatchSize(Tw.size());
//	scalar qCFD=BL.qCFD(runTime.value()); // if modified, to be modified in setMotion as welll
	scalar Hr=BL.Hr(runTime.value());
	scalar qrad=BL.qrad(runTime.value());
	scalarField BprimeCwBF = BprimeCw.boundaryField()[movingPatchID];
	scalarField mDotGywBF = mDotGface.boundaryField()[movingPatchID]& (-nf);
	scalarField hwBF = hw.boundaryField()[movingPatchID];
	scalarField hgBF = hg.boundaryField()[movingPatchID];
	scalarField fluxMapw = fluxMap.boundaryField()[movingPatchID];
	scalarField rhoeUeCH=BL.rhoeUeCH(runTime.value())*fluxMapw; // if modified, to be modified in setMotion as well
	scalar Tbackground = BL.Tinf(runTime.value());
	scalar lambda=BL.lambda(runTime.value());
	tau_w = tau.boundaryField()[movingPatchID];
	scalarField rho_sw = rho_s.boundaryField()[movingPatchID];
	scalarField Twint = Ta.boundaryField()[movingPatchID].patchInternalField();
	scalarField invDxw = mesh.deltaCoeffs().boundaryField()[movingPatchID];

// CMA temperature boundary condition used during the heating pulse; for Ref. see NASA CR 1060 (Kendall et al.; 1968)
if ((BL.heatOn(runTime.value())==1)) 
{
	forAll (TbcPatchSize, i)
	{
// Blowing correction
if ((mDotGywBF[i]<=1e-6) | (rhoeUeCH[i]<=1e-4)) {BlowingCorrection.boundaryField()[movingPatchID][i]= 1.0;}
else
{BlowingCorrection.boundaryField()[movingPatchID][i]= 
::log(1+2*lambda*(mDotGywBF[i]/(rhoeUeCH[i]*BlowingCorrection.boundaryField()[movingPatchID][i])+BprimeCwBF[i]))
/(2*lambda*(mDotGywBF[i]/(rhoeUeCH[i]*BlowingCorrection.boundaryField()[movingPatchID][i])+BprimeCwBF[i]));}

BprimeGw.boundaryField()[movingPatchID][i] = mDotGywBF[i]/(rhoeUeCH[i]*BlowingCorrection.boundaryField()[movingPatchID][i]);

// Temperature gradient at the wall (in the solid)
 	Ta.boundaryField()[movingPatchID][i] ==  Twint[i] +
		( 
			1/ (kProj[i] * invDxw[i]) *
			(
			rhoeUeCH[i]
			* BlowingCorrection.boundaryField()[movingPatchID][i]
			* (Hr- hwBF[i])
			+ mDotGywBF[i] * (hgBF[i]-hwBF[i])
			+ BprimeCwBF[i]*rhoeUeCH[i]*BlowingCorrection.boundaryField()[movingPatchID][i]*(charred.h(pw[i],Tw[i])-hwBF[i])
			- (virgin.eps(p[i], Tw[i]) * tau_w[i] * rho_v/rho_sw[i] + charred.eps(p[i], Tw[i]) * (1-tau_w[i]* rho_v/rho_sw[i])) * sigmaPlanck.value()* (pow4(Tw[i]) - pow4(Tbackground))
			+ (virgin.alpha(pw[i], Tw[i]) * tau_w[i]* rho_v/rho_sw[i] + charred.alpha(pw[i], Tw[i])* (1-tau_w[i]* rho_v/rho_sw[i])) * qrad
			)
		);
	}
} // end of if
else   // basic re-radiation boundary condition used for cool-down
{
forAll (TbcPatchSize, i)
	{
 	Tbc.boundaryField()[movingPatchID][i] = 
		( 
			1/ kProj[i] * 
			(		
//			- mDotGywBF[i]* (hwBF[i]-hgBF[i])
			- (virgin.eps(pw[i], Tw[i]) * tau_w[i] * rho_v/rho_sw[i] + charred.eps(pw[i], Tw[i]) * (1-tau_w[i]* rho_v/rho_sw[i])) * sigmaPlanck.value()* ( pow4(Tw[i]) - pow4(Tbackground) )
			+ (virgin.alpha(pw[i], Tw[i]) * tau_w[i] * rho_v/rho_sw[i] + charred.alpha(pw[i], Tw[i])* (1-tau_w[i]* rho_v/rho_sw[i])) * qrad
			)
		);
	}

} // end of else */
} // end of "if boundaryLayerApprox = "yes" (see description in the header)"
else // if boundaryLayerApprox = "no" (see description in the header)
{
	// restricts to the moving part of the mesh
   	volScalarField& Tfield =
        const_cast<volScalarField&>
        (
            motionMesh.objectRegistry::lookupObject<volScalarField>
            (
                "Ta"
            )
        );

	// access to the right patch of the moving mesh
    	fixedValueFvPatchScalarField& MovingPatchT =
        refCast<fixedValueFvPatchScalarField>
        (
            Tfield.boundaryFieldRef()[movingPatchID]
        );

	// modification of the value
	scalarField TpatchSize(MovingPatchT.size()); 

	// updating BL values
	scalar Tw=BL.Tw(runTime.value());
	scalarField fluxMapw = fluxMap.boundaryField()[movingPatchID];

	forAll (TpatchSize, i)
	{
 	MovingPatchT[i] = max(Tw*fluxMapw[i], BL.Tinf(runTime.value()));
	}
} // end of "if boundaryLayerApprox = "no" (see description in the header)"

