// This is to initialize the dynamic mesh motion -- identification of the moving
// patch (receding surface) and projection of the FV mesh on a tetrahedral mesh
// for FEM resolution of elastic mesh motion.

Info << "Reading Motion Properties" << endl;

// Open the dictionary
IOdictionary dynamicMeshDict
(
    IOobject
    (
        "dynamicMeshDict",
        runTime.constant(),
        mesh,
        IOobject::MUST_READ,
        IOobject::NO_WRITE
    )
);

// read a few switches and some data
Switch meshMotion(dynamicMeshDict.lookup("meshMotion"));
Switch boundaryLayerApprox(setCase.lookup("boundaryLayerApprox"));
Switch axiSymmetricalFlux(setCase.lookup("axiSymmetricalFlux"));
Switch oneDoutput(dynamicMeshDict.lookup("oneDoutput"));
Switch plotProfiles(dynamicMeshDict.lookup("plotProfiles"));
Switch plotShape(dynamicMeshDict.lookup("plotShape"));
dimensionedVector v0(dynamicMeshDict.lookup("v0"));
dimensionedScalar fluxFactorThreshold(dynamicMeshDict.lookup("fluxFactorThreshold"));
scalar fluxThreshold = fluxFactorThreshold.value();
word movingPatchName(dynamicMeshDict.lookup("movingPatch"));
word exitPatchName(dynamicMeshDict.lookup("exitPatch"));
vector pointMotionDirection(dynamicMeshDict.lookup("pointMotionDirection"));
Switch developper(dynamicMeshDict.lookup("developper"));
Switch plotMdotGmag = dynamicMeshDict.lookupOrDefault<Switch>("plotMdotGmag", "no");
vector fluxFactorNormal(setCase.lookup("fluxFactorNormal"));
vector fluxFactorCenter(setCase.lookup("fluxFactorCenter"));
Switch fluxFactorProjection(setCase.lookup("fluxFactorProjection"));
Switch mutationTT(setCase.lookup("mutationTT"));
Switch mutationBprime(setCase.lookup("mutationBprime"));
Switch finiteRate(setCase.lookup("finiteRate"));
Switch multiComponent(setCase.lookup("multiComponent"));
Switch volAblation(setCase.lookup("volAblation"));
Switch elementConservation(setCase.lookup("elementConservation"));
scalar failureFraction = setCase.lookupOrDefault<scalar>("failureFraction", 0.0);
scalar erosionRate = setCase.lookupOrDefault<scalar>("erosionRate", 0.0);
int physicsBasedErosionModel = setCase.lookupOrDefault<int>("physicsBasedErosionModel", 0);
scalar wallSpeciesDiffusion = setCase.lookupOrDefault<scalar>("wallSpeciesDiffusion", 1);
Switch detailedSolidEnthalpies = setCase.lookupOrDefault<Switch>("detailedSolidEnthalpies", "no");
Switch verifyThermalEquilibrium = setCase.lookupOrDefault<Switch>("verifyThermalEquilibrium", "no");
Switch constrainedEquilibrium = setCase.lookupOrDefault<Switch>("constrainedEquilibrium", "no");
Switch Darcy = setCase.lookupOrDefault<Switch>("Darcy", "no");

// Policing option compatibilities
if (volAblation && !finiteRate && !developper)
{
    Info << "WARNING: Finite-rate volume ablation (volAblation) is not compatible with equilibrium chemistry. Exiting PAM." << nl;
    return (0);
}

// Policing option compatibilities
if (!mutationTT && finiteRate && !developper)
{
    Info << "WARNING: Finite-rate volume ablation needs to be run with mutationTT. We cannot rely on tables. Exiting PAM." << nl;
    return (0);
}

// prepare the structure for the elastic motion solver

volVectorField Motionw
(
    IOobject
    (
        "Motionw",
        runTime.timeName(),
        mesh,
        IOobject::NO_READ,
        IOobject::NO_WRITE
    ),
    un * v0
);

label movingPatchID = mesh.boundaryMesh().findPatchID(movingPatchName);
label exitPatchID   = mesh.boundaryMesh().findPatchID(exitPatchName);

if (movingPatchID < 0)
{
    FatalErrorIn(args.executable())
        << "The moving patch name has not been found. "
            "It needs to be defined - even if the motion is set to zero. "
            "Please check input file 'dynamicMeshDict'."
        << abort(FatalError);
}

// Grab motion field
// Read moving patch data
word movingRegionName(dynamicMeshDict.lookup("movingRegion"));

const fvMesh &motionMesh =
    runTime.objectRegistry::lookupObject<fvMesh>(movingRegionName);

volVectorField cellMotionU
(
    IOobject
    (
        "cellMotionU",
        runTime.timeName(),
        mesh,
        IOobject::MUST_READ,
        IOobject::AUTO_WRITE
    ),
    mesh
);

volVectorField& motionU =
    const_cast<volVectorField&>
    (
        motionMesh.objectRegistry::lookupObject<volVectorField>
        (
            "cellMotionU"
        )
    );

scalarField tau_w = tau.boundaryField()[movingPatchID];

if (!boundaryLayerApprox && readBoundaryTecplotTable)
{
    Info << "boundaryLayer has to be true with readBoundaryTecplotTable = true." << endl;
    return (0);
}
