/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright held by original author
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

\*---------------------------------------------------------------------------*/

#include "BLinput.H"


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::BLinput::BLinput
(
    RectangularMatrix<scalar>& BLinput2DTable,
    const bool read3DTable,
    Time& runTime,
    PtrList<dynamicFvMesh>& solidRegions
)
    :
    BLinput2DTable_(BLinput2DTable),
    indexTime_i_(0),
    indexTime_iPlus1_(0),
    currentTimeIndex_(0.0),
    f_t_(0.0),
    pw_(0.0),
    Tw_(0.0),
    v_abl_(0.0),
    rhoeUeCH_(0.0),
    Hr_(0.0),
    qrad_(0.0),
    qCFD_(0.0),
    heatOn_(0.0),
    lambda_(0.0),
    Tinf_(0.0),
    pressureFactorOn_(0.0),
    p_dyn_(0.0),
    read3DTable_(read3DTable),
    runTime_(runTime),
    currentTime_(0.0),
    solidRegions_(solidRegions)
{
    Info << "New BLinput, read3DTable = " <<  read3DTable_ << endl;
    if (read3DTable_)
    {

        IOdictionary tecplotTablesIO_
        (
            IOobject
            (
                "constant/tecplotTable",
                runTime,
                IOobject::MUST_READ,
                IOobject::NO_WRITE,
                false
            )
        );

        dictionary tecplotTable_ = tecplotTablesIO_.subDict("tecplotTable");

        fileNameList dirEntries(tecplotTable_.lookup("files"));
        List<scalar> timeEntries(tecplotTable_.lookup("times"));
        int nVariables = readScalar(tecplotTable_.lookup("numberVariables"));
        int nZones = readScalar(tecplotTable_.lookup("numberZones"));
        List<string> variables(tecplotTable_.lookup("variableNames"));
        List<List<List<List<scalar> > > > variable_values(tecplotTable_.lookup("variableValues"));

        dirEntries_ = dirEntries;
        timeEntries_ = timeEntries;
        nVariables_ = nVariables;
        nZones_ = nZones;
        tecplot_variables_ = variables;
        tecplot_variable_values_ = variable_values;

        Info << "Read Tecplot table" << endl;
        Info << "files: " ;
        forAll(dirEntries_, i)
        {
            Info << dirEntries_[i] << " ";
        }
        Info << endl;
        Info << "times: " ;
        forAll(timeEntries_, i)
        {
            Info << timeEntries_[i] << " ";
        }
        Info << endl;
        Info << "number of zones = " << nZones_ << " , number of variables = "  << nVariables_ << endl;
        Info << "variables: " ;
        forAll(variables, i)
        {
            Info << variables[i] << " ";
        }
        Info << endl;

        currentTimeVariables_.resize(variables.size());
        currentVariables_.resize(variables.size());

        forAll(variables, varI)
        {
            string tecplot_index = variables[varI];
            if (tecplot_index == "xw(m)") tecplotTable_indexCoordX = varI;
            if (tecplot_index == "yw(m)")tecplotTable_indexCoordY = varI;
            if (tecplot_index == "zw(m)")tecplotTable_indexCoordZ = varI;
            if (tecplot_index == "pw(Pa)")tecplotTable_indexPw = varI;
            if (tecplot_index == "rhoe(kg/m^3)")tecplotTable_indexRhoe = varI;
            if (tecplot_index == "He(J/kg)")tecplotTable_indexHe = varI;
            if (tecplot_index == "ue(m/s)")tecplotTable_indexUe = varI;
            if (tecplot_index == "ve(m/s)")tecplotTable_indexVe = varI;
            if (tecplot_index == "we(m/s)")tecplotTable_indexWe = varI;
            if (tecplot_index == "CH(kg/m^2.s)") tecplotTable_indexCH = varI;

            currentVariables_.set(varI, new PtrList<volScalarField>);
            currentVariables_[varI].resize(solidRegions.size());
            currentTimeVariables_[varI].resize(solidRegions.size());
            forAll(solidRegions, regionI)
            {
                currentVariables_[varI].set
                (
                    regionI ,
                    new volScalarField
                    (
                        IOobject
                        (
                            variables[varI],
                            runTime.timeName(),
                            solidRegions[regionI],
                            IOobject::NO_READ,
                            IOobject::NO_WRITE
                        ),
                        solidRegions[regionI],
                        dimensionSet(0, 0, 0, 0, 0)
                    )
                );

                currentTimeVariables_[varI][regionI].resize(solidRegions[regionI].boundaryMesh().size());
                forAll(solidRegions[regionI].boundaryMesh(), patchI)
                {
                    currentTimeVariables_[varI][regionI][patchI] = -1;
                }
            }
        }


        int nRows = BLinput2DTable_.m();
        int nColumns = BLinput2DTable_.n();

        for (int rowI = 0; rowI < nRows; rowI++)
        {
            for (int columnI = 0 ; columnI < nColumns; columnI++)
            {
                BLinput2DTable_[rowI][columnI] = 0;
            }
        }

        scalar lastTime = max(timeEntries);

        BLinput2DTable_[0][0] = 0;
        BLinput2DTable_[1][0] = lastTime;

        int heatOn = 1;
        BLinput2DTable_[0][8] = heatOn;
        BLinput2DTable_[1][8] = heatOn;

        scalar lambda = 0.5;
        BLinput2DTable_[0][9] = lambda;
        BLinput2DTable_[1][9] = lambda;

        scalar Tinf = 50;
        BLinput2DTable_[0][10] = Tinf;
        BLinput2DTable_[1][10] = Tinf;

        f_t_ = 1;

        regionNames_.resize(solidRegions.size());

        forAll(solidRegions, regionI)
        {
            regionNames_[regionI] = solidRegions[regionI].dbDir();
        }
    }
}

// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

Foam::BLinput::~BLinput()
{}

// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

// indexTi: Returns the Time interpolation factor and updates indexTi_ and indexTiPlus1_
Foam::scalar Foam::BLinput::update(scalar Time_)
{
    if
    (
        Time_ < BLinput2DTable_[0][0] ||
        Time_ > BLinput2DTable_[BLinput2DTable_.m() - 1][0])
    {
        Info << "Time out of B.L. conditions input table range" << nl;
        if (Time_ < BLinput2DTable_[0][0])
        {
            Info << "Time = " << Time_ << " < " << BLinput2DTable_[0][0] << nl;
        }

        if (Time_ > BLinput2DTable_[BLinput2DTable_.m() - 1][0])
        {
            Info << "Time = " << Time_
                 << " > " << BLinput2DTable_[BLinput2DTable_.m() - 1][0]
                 << nl;
        }
        ::exit(1);
    }

    int i_Time(0);
    while (BLinput2DTable_[i_Time][0] < Time_)
    {
        i_Time++;
    }

    indexTime_i_ = i_Time;

    if (Time_ == BLinput2DTable_[indexTime_i_][0])
    {
        indexTime_iPlus1_ = indexTime_i_;
        f_t_ = 1;
    }
    else
    {
        indexTime_iPlus1_ = indexTime_i_ - 1;
        f_t_ =
            (BLinput2DTable_[indexTime_iPlus1_][0] - Time_) /
            (BLinput2DTable_[indexTime_iPlus1_][0] - BLinput2DTable_[indexTime_i_][0]);
    }

    return f_t_;
    currentTimeIndex_ = Time_;
}

Foam::scalarField Foam::BLinput::update(fvMesh& mesh, label patchID, int variableI)
{
    currentTime_ = runTime_.time().value();

    if ( (currentTime_ < timeEntries_[0]) | (currentTime_ > timeEntries_[timeEntries_.size() - 1]))
    {
        Info << "Time out of B.L. conditions input table range" << nl;
        ::exit(1);
    }

    int i_Time(0);
    while (timeEntries_[i_Time] < currentTime_)
    {
        i_Time++;
    }
    indexTime_i_ = i_Time;
    if (currentTime_ == timeEntries_[indexTime_i_])
    {
        indexTime_iPlus1_ = indexTime_i_;
        f_t_ = 1;
    }
    else
    {
        indexTime_iPlus1_ = indexTime_i_ - 1;
        f_t_ = (timeEntries_[indexTime_iPlus1_] - currentTime_) / (timeEntries_[indexTime_iPlus1_] - timeEntries_[indexTime_i_]);
    }


    int currentRegion = -1;
    forAll(regionNames_, regionI)
    {
        if (mesh.dbDir() == regionNames_[regionI])
        {
            currentRegion = regionI;
        }

    }
    if (currentRegion < 0)
    {
        FatalErrorInFunction
                << "Error: " << mesh.dbDir() << " mesh is not in solidRegions."
                << exit(FatalError);
    }

    scalar time = runTime_.time().value();
    scalar tol = 1e-1;

    if ((time - currentTimeVariables_[variableI][currentRegion][patchID]) < tol && currentTimeVariables_[variableI][currentRegion][patchID] > 0)
    {
        Info << "Using previous " << tecplot_variables_[variableI] << endl;
        return currentVariables_[variableI][currentRegion].boundaryFieldRef()[patchID];
    }
    else
    {
        Info << "Updating " << tecplot_variables_[variableI] << " patchID = " << patchID <<  endl;
        currentTimeVariables_[variableI][currentRegion][patchID] = time;

        scalarField values_i = interpolateEnvironment(mesh, patchID, variableI, indexTime_i_);
        scalarField values_iPlus1 = interpolateEnvironment(mesh, patchID, variableI, indexTime_iPlus1_);
        scalarField values =  values_i * f_t_ + values_iPlus1 * (1 - f_t_);
        currentVariables_[variableI][currentRegion].boundaryFieldRef()[patchID] = values;
        return values;
    }

}



Foam::scalarField Foam::BLinput::interpolateEnvironment(fvMesh& mesh, label patchID, int variableI, int indexTime)
{
    scalarField field_environment(mesh.boundaryMesh()[patchID].size());

    forAll(mesh.boundaryMesh()[patchID], faceI)
    {

        scalar value_environment = 0;
        int num_i = 0;

        forAll(mesh.boundaryMesh()[patchID][faceI], pI)
        {

            vector pointI = mesh.points()[mesh.boundaryMesh()[patchID][faceI][pI]];
            double xp = pointI[0];
            double yp = pointI[1];
            double zp = pointI[2];

            scalar distance_min = -1;
            label indexTecplotTable = -1;
            label zoneTecplotTable = -1;

            forAll(tecplot_variable_values_[indexTime], zoneI)
            {
                forAll(tecplot_variable_values_[indexTime][zoneI][0], i)
                {
                    double x = tecplot_variable_values_[indexTime][zoneI][0][i];
                    double y = tecplot_variable_values_[indexTime][zoneI][1][i];
                    double z = tecplot_variable_values_[indexTime][zoneI][2][i];
                    double y_symm = -y;

                    double dist = ::sqrt(::pow(x - xp, 2) + ::pow(y - yp, 2) + ::pow(z - zp, 2)) ;
                    double dist_symm = ::sqrt(::pow(x - xp, 2) + ::pow(y_symm - yp, 2) + ::pow(z - zp, 2)) ;
                    if (dist_symm < dist)
                    {
                        dist = dist_symm;
                    }
                    if (distance_min < 0 || dist < distance_min)
                    {
                        distance_min = dist;
                        indexTecplotTable = i;
                        zoneTecplotTable = zoneI;
                    }
                }
            }

            if (indexTecplotTable >= 0 && zoneTecplotTable >= 0)
            {
                value_environment += tecplot_variable_values_[indexTime][zoneTecplotTable][variableI][indexTecplotTable];
                num_i++;
            }

        }
        if (num_i != 0)
        {
            value_environment /= num_i;
        }

        field_environment[faceI] = value_environment;

    }

    return field_environment;

}


Foam::scalar Foam::BLinput::pw(scalar Time_)
{
    if (currentTimeIndex_ != Time_)
    {
        update(Time_);
    }

    return pw();
}

Foam::scalarField Foam::BLinput::pw(fvMesh& mesh, label patchID)
{
    if (read3DTable_)
    {

        scalarField pw = update(mesh, patchID, tecplotTable_indexPw);

        scalar pw_min = 10; // Pa
        forAll(pw, faceI)
        {
            if (pw[faceI] < pw_min)
            {
                pw[faceI] = pw_min;
            }
        }
        return pw;
    }
    else
    {
        scalarField zero(1);
        zero[1] = 0;
        return zero;
    }
}

Foam::scalar Foam::BLinput::Tw(scalar Time_)
{
    if (currentTimeIndex_ != Time_)
    {
        update(Time_);
    }

    return Tw();
}

Foam::scalar Foam::BLinput::rhoeUeCH(scalar Time_)
{
    if (currentTimeIndex_ != Time_)
    {
        update(Time_);
    }

    return rhoeUeCH();
}


Foam::scalarField Foam::BLinput::rhoeUeCH(fvMesh& mesh, label patchID)
{
    if (read3DTable_)
    {
        //scalarField rhoe =update(mesh,patchID,tecplotTable_indexRhoe);
        //scalarField ue =update(mesh,patchID,tecplotTable_indexUe);
        // scalarField ve =update(mesh,patchID,tecplotTable_indexVe);
        //       scalarField we =update(mesh,patchID,tecplotTable_indexWe);
        scalarField CH = update(mesh, patchID, tecplotTable_indexCH);

        // vectorField nf =   - mesh.Sf().boundaryField()[patchID]/mesh.magSf().boundaryField()[patchID];

        scalarField rhoeUeCH(CH.size());
        forAll(rhoeUeCH, faceI)
        {
            //vector Ue(ue[faceI],ve[faceI],we[faceI]);
            scalar rhoeUeCH_faceI = CH[faceI]; //*mag(Ue)*rhoe[faceI];
            rhoeUeCH[faceI] = rhoeUeCH_faceI;
        }

        return rhoeUeCH;
    }
    else
    {
        scalarField zero(1);
        zero[1] = 0;
        return zero;
    }
}

Foam::scalar Foam::BLinput::Hr(scalar Time_)
{
    if (currentTimeIndex_ != Time_)
    {
        update(Time_);
    }

    return Hr();
}

Foam::scalarField Foam::BLinput::Hr(fvMesh& mesh, label patchID)
{
    if (read3DTable_)
    {
        scalarField Hr = update(mesh, patchID, tecplotTable_indexHe);
        return Hr;
    }
    else
    {
        scalarField zero(1);
        zero[1] = 0;
        return zero;
    }
}


Foam::scalar Foam::BLinput::qrad(scalar Time_)
{
    if (currentTimeIndex_ != Time_)
    {
        update(Time_);
    }

    return qrad();
}

Foam::scalar Foam::BLinput::qCFD(scalar Time_)
{
    if (currentTimeIndex_ != Time_)
    {
        update(Time_);
    }

    return qCFD();
}

Foam::scalar Foam::BLinput::heatOn(scalar Time_)
{
    if (currentTimeIndex_ != Time_)
    {
        update(Time_);
    }

    return heatOn();
}

Foam::scalar Foam::BLinput::lambda(scalar Time_)
{
    if (currentTimeIndex_ != Time_)
    {
        update(Time_);
    }

    return lambda();
}

Foam::scalar Foam::BLinput::Tinf(scalar Time_)
{
    if (currentTimeIndex_ != Time_)
    {
        update(Time_);
    }

    return Tinf();
}

Foam::scalar Foam::BLinput::pressureFactorOn(scalar Time_)
{
    if (currentTimeIndex_ != Time_)
    {
        update(Time_);
    }

    return pressureFactorOn();
}

Foam::scalar Foam::BLinput::p_dyn(scalar Time_)
{
    if (currentTimeIndex_ != Time_)
    {
        update(Time_);
    }

    return p_dyn();
}

// * * * * * * * * * * * * * * * END Member Functions  * * * * * * * * * * * * * //
